use crate::schema;
use microrm::prelude::*;

mod oidc;
mod session;

pub struct ServerCoreState {
    pool: microrm::DBPool<'static>,
    templates: handlebars::Handlebars<'static>,
}

#[derive(Clone)]
struct ServerState {
    core: &'static ServerCoreState,
}

fn is_auth_valid<T>(core: &'static ServerCoreState, of: &tide::Request<T>) -> Option<bool> {
    let cookie = of.cookie("vogt_session")?;
    let session_id = cookie.value();

    Some(
        core.pool
            .query_interface()
            .get()
            .by(schema::Session::Key, &session_id)
            .one()
            .expect("couldn't query db")
            .is_some(),
    )
}

pub async fn run_server(db: microrm::DB, port: u16) {
    let db_box = Box::new(db);
    let db: &'static mut microrm::DB = Box::leak(db_box);
    let pool = microrm::DBPool::new(db);

    let core_state = Box::leak(Box::new(ServerCoreState {
        pool,
        templates: handlebars::Handlebars::new(),
    }));

    // XXX: for development only
    // core_state.templates.write().unwrap().set_dev_mode(true);

    core_state
        .templates
        .register_templates_directory(".tmpl", "tmpl/")
        .expect("Couldn't open templates directory?");
    println!("registered templates:");
    for tmpl in core_state.templates.get_templates() {
        println!("- {}", tmpl.0);
    }

    core_state.templates.render("id_v1_login", &()).unwrap();

    let state = ServerState { core: core_state };

    let mut app = tide::with_state(state);

    app.with(tide::log::LogMiddleware::new());

    app.at("/static")
        .serve_dir("static/")
        .expect("Can't serve static files");

    app.at("/:realm/v1/session/")
        .nest(session::session_v1_server(core_state));
    app.at("/:realm/v1/oidc/")
        .nest(oidc::oidc_v1_server(core_state));

    app.listen(("127.0.0.1", port)).await.expect("Can listen");
}
