use crate::{schema, user};
use microrm::prelude::*;
use serde::Deserialize;
use tide::http::Cookie;

#[derive(Clone)]
pub struct ServerState {
    core: &'static super::ServerCoreState,
    realm_cache:
        std::sync::Arc<std::sync::RwLock<std::collections::HashMap<String, schema::RealmID>>>,
}

type Request = tide::Request<ServerState>;

impl ServerState {
    pub fn get_realm(&self, req: &Request) -> Option<schema::RealmID> {
        let realm_str = req
            .param("realm")
            .expect("get_realm called with no :realm param");
        let cache = self.realm_cache.read().unwrap();
        let cache_lookup = cache.get(realm_str);

        // expected case
        if cache_lookup.is_some() {
            return cache_lookup.map(|x| *x);
        }
        drop(cache);

        // unexpected case, but maybe we haven't filled that cache entry yet

        let qi = self.core.pool.query_interface();
        let realm = qi
            .get()
            .by(schema::Realm::Shortname, realm_str)
            .one()
            .expect("couldn't query db");

        if let Some(with_id) = realm {
            let mut cache = self.realm_cache.write().unwrap();
            cache.insert(realm_str.to_owned(), with_id.id());
            return Some(with_id.id());
        }

        // other expected case, is bogus realm
        return None;
    }

    fn build_session(
        qi: &microrm::QueryInterface,
    ) -> tide::Result<(schema::SessionID, Option<tide::http::Cookie<'static>>)> {
        let rng = ring::rand::SystemRandom::new();
        let session_id: [u8; 32] = ring::rand::generate(&rng)
            .map_err(|_| tide::Error::from_str(500, "Failed to generate session ID"))?
            .expose();
        let session_id = base64::encode_config(session_id, base64::URL_SAFE_NO_PAD);

        let maybe_id = qi.add(&schema::Session {
            key: session_id.clone(),
        });
        Ok((
            maybe_id.ok().ok_or(tide::Error::from_str(
                500,
                "Failed to store session in database",
            ))?,
            Some(Cookie::new("vogt_session", session_id)),
        ))
    }

    pub fn verify_session(&self, req: &Request) -> Option<(schema::RealmID, schema::UserID)> {
        self.get_or_build_session(req)
            .ok()
            .zip(self.get_realm(req))
            .and_then(|((sid, _cookie), realm)| {
                self.get_auth_for_session(realm, sid).and_then(|auth| {
                    if auth.challenges_left.len() == 0 {
                        Some((realm, auth.user))
                    } else {
                        None
                    }
                })
            })
    }

    pub fn get_or_build_session(
        &self,
        req: &Request,
    ) -> tide::Result<(schema::SessionID, Option<tide::http::Cookie<'static>>)> {
        let qi = self.core.pool.query_interface();
        if let Some(sid) = req.cookie("vogt_session") {
            let existing = qi
                .get()
                .by(schema::Session::Key, sid.value())
                .one()
                .expect("couldn't query db");

            if existing.is_some() {
                return Ok((existing.unwrap().id(), None));
            }
        }
        Self::build_session(qi)
    }

    pub fn get_auth_for_session(
        &self,
        realm: schema::RealmID,
        session: schema::SessionID,
    ) -> Option<microrm::WithID<schema::SessionAuthentication>> {
        let qi = self.core.pool.query_interface();

        use schema::SessionAuthentication as SAC;
        qi.get()
            .by(SAC::Realm, &realm)
            .by(SAC::Session, &session)
            .one()
            .expect("couldn't query db")
    }

    pub fn destroy_auth(&self, realm: schema::RealmID, session: schema::SessionID) {
        let qi = self.core.pool.query_interface();

        use schema::SessionAuthentication as SAC;
        qi.delete()
            .by(SAC::Realm, &realm)
            .by(SAC::Session, &session)
            .exec()
            .expect("couldn't query db")
    }
}

impl ServerState {
    fn render_login_from_auth(
        &self,
        mut response: tide::Response,
        auth: Option<schema::SessionAuthentication>,
        error_msg: Option<String>,
    ) -> tide::Response {
        log::info!("rendering login response... auth is {:?}", auth);

        let to_present: Option<schema::AuthChallengeType> = match auth {
            None => Some(schema::AuthChallengeType::Username),
            Some(auth) => auth.challenges_left.first().copied(),
        };

        if to_present.is_none() {
            response.set_status(302);
            tide::Redirect::new("/").into()
        } else {
            self.render_login_page(response, to_present.unwrap(), error_msg)
        }
    }

    fn render_login_page(
        &self,
        mut response: tide::Response,
        to_present: schema::AuthChallengeType,
        error_msg: Option<String>,
    ) -> tide::Response {
        let tmpl = &self.core.templates;

        let do_challenge = |ty, ch| {
            tmpl.render(
                "id_v1_login",
                &serde_json::json!(
                    {
                        "challenge":
                            format!(r#"
                            <input type="hidden" name="challenge_type" value="{:?}" />
                            <div class="challenge-type">{}</div>
                            <div class="challenge-content">{}</div>
                            "#,
                                to_present, ty, ch),
                        "error_msg": error_msg.iter().collect::<Vec<_>>()
                    }
                ),
            )
            .unwrap()
        };

        response.set_content_type("text/html");

        match to_present {
            schema::AuthChallengeType::Username => {
                response.set_body(do_challenge(
                    "Username",
                    r#"<input name="challenge" type="text" autofocus />"#,
                ));
            }
            schema::AuthChallengeType::Password => {
                response.set_body(do_challenge(
                    "Password",
                    r#"<input name="challenge" type="password" autofocus />"#,
                ));
            }
            _ => todo!(),
        }

        response
    }
}

async fn v1_login(req: tide::Request<ServerState>) -> tide::Result<tide::Response> {
    let mut response = tide::Response::builder(200).build();

    let realm = req
        .state()
        .get_realm(&req)
        .ok_or(tide::Error::from_str(404, "No such realm"))?;
    let (session_id, cookie) = req.state().get_or_build_session(&req)?;
    cookie.map(|c| response.insert_cookie(c));

    let auth = req.state().get_auth_for_session(realm, session_id);

    Ok(req
        .state()
        .render_login_from_auth(response, auth.map(|a| a.wrapped()), None))
}

async fn v1_login_post(mut req: tide::Request<ServerState>) -> tide::Result<tide::Response> {
    let mut response = tide::Response::builder(200).build();

    let realm = req
        .state()
        .get_realm(&req)
        .ok_or(tide::Error::from_str(404, "No such realm"))?;
    let (session_id, cookie) = req.state().get_or_build_session(&req)?;
    cookie.map(|c| response.insert_cookie(c));

    let mut auth = req.state().get_auth_for_session(realm, session_id);

    #[derive(Deserialize)]
    struct ResponseBody {
        challenge_type: String,
        challenge: String,
        reset: Option<String>,
    }

    let body: ResponseBody = req.body_form().await?;

    // check if a login reset was requested; if so, we start again from the top
    if body.reset.is_some() {
        if let Some(_) = auth {
            req.state().destroy_auth(realm, session_id);
            response.set_status(302);
            return Ok(tide::Redirect::new("login").into());
        }
    }

    use schema::AuthChallengeType as ChallengeType;

    let challenge: schema::AuthChallengeType = match body.challenge_type.as_str() {
        "Username" => ChallengeType::Username,
        "Password" => ChallengeType::Password,
        _ => Err(tide::Error::from_str(400, "Unknown challenge type"))?,
    };

    let mut error = None;

    // check that the response matches what we're expecting next
    let to_be_presented: Option<schema::AuthChallengeType> = match &auth {
        None => Some(schema::AuthChallengeType::Username),
        Some(auth) => auth.challenges_left.first().copied(),
    };

    if to_be_presented != Some(challenge) {
        Err(tide::Error::from_str(400, "Unexpected challenge type"))?
    }

    match challenge {
        ChallengeType::Username => {
            let qi = req.state().core.pool.query_interface();
            req.state().destroy_auth(realm, session_id);

            let user = qi
                .get()
                .by(schema::User::Realm, &realm)
                .by(schema::User::Username, &body.challenge)
                .one()
                .expect("couldn't query db");
            if user.is_none() {
                error = Some(format!("No such user {}", body.challenge));
            } else {
                let user = user.unwrap();

                // TODO: set list of challenges to be whatever else this user has set up
                let sa = schema::SessionAuthentication {
                    session: session_id,
                    realm: realm,
                    user: user.id(),
                    challenges_left: vec![schema::AuthChallengeType::Password],
                };
                let id = qi.add(&sa).unwrap();
                auth = Some(microrm::WithID::new(sa, id));
            }
        }
        ct => {
            if let Some(auth) = &mut auth {
                let qi = req.state().core.pool.query_interface();

                let user = qi.get().by_id(&auth.user).one().expect("couldn't query db");

                if let Some(user) = user {
                    let user = user::User::from_model(user);

                    let verification = user.verify_challenge(&qi, ct, body.challenge.as_bytes());

                    match verification {
                        Some(true) => {
                            auth.challenges_left.remove(0);
                            qi.update()
                                .to(auth.as_ref())
                                .by_id(&auth.id())
                                .exec()
                                .expect("couldn't update auth status?");
                        }
                        Some(false) => {
                            error = Some("Incorrect response. Please try again".into());
                        }
                        None => {
                            error = Some(
                                "User no longer exists. Please contact an administrator.".into(),
                            );
                        }
                    }
                } else {
                    error = Some(format!("User is not configured correctly: either it was deleted or it lacks a required authentication challenge type. Please contact an administrator."));
                }
            } else {
                error = Some(format!("Please restart login process."));
            }
        }
    };

    Ok(req
        .state()
        .render_login_from_auth(response, auth.map(|a| a.wrapped()), error))
}

async fn v1_logout(req: tide::Request<ServerState>) -> tide::Result<tide::Response> {
    let realm = req
        .state()
        .get_realm(&req)
        .ok_or(tide::Error::from_str(404, "No such realm"))?;
    let (session_id, _) = req.state().get_or_build_session(&req)?;

    req.state().destroy_auth(realm, session_id);
    Ok(tide::Redirect::new("/").into())
}

pub fn session_v1_server(core: &'static super::ServerCoreState) -> tide::Server<ServerState> {
    let mut srv = tide::with_state(ServerState {
        core,
        realm_cache: std::sync::Arc::new(std::sync::RwLock::new(std::collections::HashMap::new())),
    });

    srv.with(tide::log::LogMiddleware::new());

    srv.at("login").get(v1_login).post(v1_login_post);
    srv.at("logout").get(v1_logout);

    srv
}
